package clustering;

import java.io.File;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.List;

import clustering.node.Folder;
import clustering.node.Leaf;

import com.giantleap.common.nodebean.node.base.filter.AndFilter;
import com.giantleap.common.nodebean.node.base.filter.DoubleAttributeFilter;
import com.giantleap.common.nodebean.node.base.filter.HasParentFilter;
import com.giantleap.common.nodebean.node.base.filter.NodeClassFilter;
import com.giantleap.common.nodebean.node.base.storage.node.Node;
import com.giantleap.common.nodebean.node.root.file.FileRootNode;
import com.giantleap.common.nodebean.persistence.JDOM.JDOMPersistenceMgr;

public class NewClonalRelate {

	public static void main(String[] args) throws Exception {
		Date date1 = new Date();
		long t1 = date1.getTime();
		NewClonalRelate newclonal = new NewClonalRelate();
		newclonal = newclonal.ClonalRelate(newclonal);
		Date date2 = new Date();
		long t2 = date2.getTime();
		System.out.print(t2-t1);
		System.out.println("ms");
	}

	public NewClonalRelate ClonalRelate(NewClonalRelate newclonal) throws Exception {
		String inputFilename = InOutFileName.getInputFilename();
		NewFindCDR findcdr = new NewFindCDR();
		ArrayList<CDR3> cdr3s = findcdr.getCDR3s(inputFilename);
		genTest(inputFilename, cdr3s);
		readTest(inputFilename, cdr3s);
		return newclonal;
	}

	public void genTest(String inputFilename, ArrayList<CDR3> cdr3s) {
		Float[][] distanceMatrix = newCalculateDistanceMatrix(cdr3s);
		String outFilename = inputFilename.substring(0, inputFilename
				.lastIndexOf("."));
		//System.out.println(outFilename);
		newgetTree(distanceMatrix, outFilename + ".dendrogram.xml", cdr3s);
		NewickConverter newick = new NewickConverter();
		newick.converter(outFilename + ".dendrogram.xml");
	}

	public Float[][] newCalculateDistanceMatrix(ArrayList<CDR3> cdr3s) {

		Float[][] distanceMatrix = new Float[cdr3s.size() - 1][];
		// System.out.println("col size ="+ distanceMatrix.length);
		DistanceMeasure measure = new DistanceMeasure();
		//double total_length = 0;
		for (int i = 0; i < cdr3s.size() - 1; i++) {
			CDR3 cdr3 = cdr3s.get(i);
			//System.out.println(cdr3.getSeqID());
			Float[] distance_matrix_row = new Float[cdr3s.size() - i - 1];
			String row = cdr3.getCDR3();
			String vrow = cdr3.getIGHVName();
			String jrow = cdr3.getIGHJName();
			for (int j = i + 1; j < cdr3s.size(); j++) {
				// System.out.println(j);
				// System.out.println(names.get(j));
				CDR3 cdr3_1 = cdr3s.get(j);
				String col = cdr3_1.getCDR3();
				String vcol = cdr3_1.getIGHVName();
				String jcol = cdr3_1.getIGHJName();
				int[] disVal = measure.NED(row, col);
				float cell_value = (float) (disVal[0]
						+ measure.VMatrix(vrow, vcol) + measure.JMatrix(jrow,
						jcol))
						/ disVal[1];
				distance_matrix_row[j - i - 1] = cell_value;
			}
			distanceMatrix[i] = distance_matrix_row;
			//total_length+= cdr3.getCDR3length();
		}
		//total_length+=cdr3s.get(cdr3s.size()-1).getCDR3length();
		//total_length = total_length/cdr3s.size();
		//System.out.println(total_length);
		return distanceMatrix;
	}

	public boolean newgetTree(Float[][] distanceMatrix, String outXMLTree,
			ArrayList<CDR3> cdr3s) {
		// System.out.println(distanceMatrix[0].length);
		Folder folder = UPGMATreeGenerator.newgenerateTree(distanceMatrix,
				cdr3s);
		FileRootNode rootNode = new FileRootNode("root");
		rootNode.makeLogicRoot();
		rootNode.addChildNode(folder);
		JDOMPersistenceMgr mgr = JDOMPersistenceMgr.getInstance();
		return mgr.save(rootNode, outXMLTree);

	}

	public Folder loadTree(String filename) {
		if (filename == null)
			throw new IllegalArgumentException();
		File file = new File(filename);
		if (!file.exists())
			throw new RuntimeException();
		JDOMPersistenceMgr mgr = JDOMPersistenceMgr.getInstance();
		Node node = mgr.loadStandalone(filename);
		Folder folder = null;
		if (node != null) {
			Collection<Node> childList = node.getAllChildNode();
			if (childList.size() > 0) {
				Node[] nodeArray = childList.toArray(new Node[0]);
				Node folderNode = nodeArray[0];
				if (folderNode instanceof Folder) {
					folder = (Folder) folderNode;
				}
			}
		}
		return folder;
	}

	public List<Folder> cutFolder(Folder folder, float distance) {
		if (folder == null)
			throw new IllegalArgumentException();
		DoubleAttributeFilter parentDistFilter = new DoubleAttributeFilter(
				"_dist", distance, "GT");
		DoubleAttributeFilter distFilter = new DoubleAttributeFilter("_dist",
				distance, "LS_OR_EQ");
		HasParentFilter pFilter = new HasParentFilter(parentDistFilter, false);
		NodeClassFilter classFilter = new NodeClassFilter(Folder.NODE_TYPE);

		AndFilter nodefilter = new AndFilter(classFilter, distFilter);
		AndFilter filter = new AndFilter(nodefilter, pFilter);

		List<Node> nodes = folder.extract(filter, true);
		List<Folder> folders = new ArrayList<Folder>();
		for (Node node : nodes) {
			if (node instanceof Folder) {
				Folder f = (Folder) node;
				folders.add(f);
			} else {
				throw new RuntimeException();
			}
		}
		return folders;
	}

	public void sortCluster(List<SortCluster> clusters) {
		if (clusters == null)
			throw new IllegalArgumentException();
		Collections.sort(clusters, SortCluster.getClusterIdComparator());
	}

	public void readTest(String inputFilename, ArrayList<CDR3> cdr3s)
			throws Exception {

		String xmlpath = inputFilename.substring(0, inputFilename
				.lastIndexOf("."))
				+ ".dendrogram.xml";
		Folder folder = loadTree(xmlpath);

		// Use the below code to get maximum disance of a dataset.
		/*
		 * 
		 * List<String> ids = new ArrayList<String>();
		 * 
		 * //ids.add("5"); ids.add("27"); ids.add("20"); ids.add("19"); for (int
		 * i = 0; i < 90; i++) { // Integer PW57 = seqs.size() - 20 + i;
		 * ids.add(String.valueOf(seqs.size() - 90 + i)); } // Get the distance
		 * that needs to perform cutting float dis = getMaxDistance(folder,
		 * ids); System.out.println(dis);
		 */

		// Distance threshold can be changed here.
		float dis = 0.32f;

		List<Folder> subtrees = cutFolder(folder, dis);
		File path = new File(xmlpath.substring(0, xmlpath.lastIndexOf("."))
				+ ".subtrees");
		if (!path.exists())
			path.mkdir();
		String pathString = path.getAbsolutePath();
		StringBuffer sb = new StringBuffer();
		List<SortCluster> clusters = new ArrayList<SortCluster>();
		HashMap<String, Integer> seq_ids = new HashMap<String, Integer>();
		for (int i = 0; i < cdr3s.size(); i++) {
			seq_ids.put(cdr3s.get(i).getSeqID(), i);
		}
		// System.out.println(seq_ids);
		for (int i = 0; i < subtrees.size(); i++) {
			Folder f = subtrees.get(i);
			NodeClassFilter classFilter = new NodeClassFilter(Leaf.NODE_TYPE);
			List<Node> leafnodes = f.extract(classFilter, true);
			List<String> seqs = new ArrayList<String>();
			for (Node leafnode : leafnodes) {
				if (leafnode instanceof Leaf) {
					Leaf leaf = (Leaf) leafnode;
					String leafId = leaf.get_id();
					// System.out.println(leafId);
					if (leafId != null) {
						try {
							// seq = Integer.parseInt(leafId);
							seqs.add(leafId);
						} catch (Exception e) {
							e.printStackTrace();
						}
					}
				}
			}
			// System.out.println(seqs);
			for (int n = 0; n < seqs.size(); n++) {
				int seqInt = seq_ids.get(seqs.get(n));
				String seqString = String.valueOf(i + 1) + ";"
						+ String.valueOf(leafnodes.size()) + ";"
						+ cdr3s.get(seqInt).getCDR3()
						+ cdr3s.get(seqInt).getCDR3length() + ";"
						+ cdr3s.get(seqInt).getOriginal();
				// + lines.get(seqInt).toString();

				Integer clusterId = i;
				SortCluster clusterResult = new SortCluster(seqString,
						clusterId);
				clusters.add(clusterResult);
			}
			// System.out.println(f.getId());
			FileRootNode rootNode = new FileRootNode("root");
			rootNode.makeLogicRoot();
			rootNode.addChildNode(f);
			JDOMPersistenceMgr mgr = JDOMPersistenceMgr.getInstance();
			String filename = pathString + "/" + String.valueOf(i + 1) + ".xml";
			// System.out.println(filename);
			mgr.save(rootNode, filename);
			NewickConverter newick = new NewickConverter();
			newick.converter(filename);
		}
		sortCluster(clusters);

		for (int i = 0; i < clusters.size(); i++) {
			SortCluster clust = clusters.get(i);
			sb.append(clust.getSequenceId());
			if (i < (clusters.size() - 1)) {
				sb.append("\n");
			}
			// System.out.println(sb);
		}
		// System.out.println(subtrees.size());
		try {
			PrintWriter writer = new PrintWriter(inputFilename.substring(0,
					inputFilename.lastIndexOf("."))
					+ ".output.txt");
			writer.print(sb.toString());
			writer.flush();
			writer.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
}
