package clustering;

import java.io.File;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Date;

import clustering.node.Folder;
import clustering.node.Leaf;

import com.giantleap.common.nodebean.node.base.filter.NodeClassFilter;
import com.giantleap.common.nodebean.node.base.storage.node.Node;
import com.giantleap.common.nodebean.node.root.file.FileRootNode;
import com.giantleap.common.nodebean.persistence.JDOM.JDOMPersistenceMgr;

public class PartitionedSetClustering {

	private static Integer cluster_id = 0;

	public static void main(String[] args) throws Exception {
		
		Date date1 = new Date();
		long t1 = date1.getTime();
		
		PartitionedSetClustering partionSets = new PartitionedSetClustering();
		partionSets = partionSets.PartionSets(partionSets);
		
		Date date2 = new Date();
		long t2 = date2.getTime();
		System.out.print(t2-t1);
		System.out.println("ms");
		
	}

	public PartitionedSetClustering PartionSets(PartitionedSetClustering partset)
			throws Exception {
		String inputFilename = InOutFileName.getInputFilename();
		NewFindCDR findcdr = new NewFindCDR();
		ArrayList<CDR3> cdr3s = findcdr.getCDR3s(inputFilename);
		ArrayList<ArrayList<CDR3>> partitioned = SelectSimilar(cdr3s);
		NewClonalRelate newclonal = new NewClonalRelate();
		StringBuffer sb = new StringBuffer();
		// Integer cluster_id = 0;
		for (int i = 0; i < partitioned.size(); i++) {
			//System.out.print(partitioned.get(i).size() +" & ");
			String update_input = inputFilename + "." + Integer.toString(i)
					+ ".txt";
			// System.out.println(update_input);
			newclonal.genTest(update_input, partitioned.get(i));
			sb = paritionedSetReadset(update_input, partitioned.get(i), sb);
			// System.out.println(cluster_id);
		}
		try {
			PrintWriter writer = new PrintWriter(inputFilename.substring(0,
					inputFilename.lastIndexOf("."))
					+ ".output.txt");
			writer.print(sb.toString());
			writer.flush();
			writer.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
		return partset;
	}

	public ArrayList<ArrayList<CDR3>> SelectSimilar(ArrayList<CDR3> cdr3s) {

		ArrayList<ArrayList<CDR3>> partitioned = new ArrayList<ArrayList<CDR3>>();
		while (cdr3s.size() > 1) {
			int li = cdr3s.get(0).getCDR3length();
			ArrayList<CDR3> partitioned_set = new ArrayList<CDR3>();
			for (int j = 1; j < cdr3s.size(); j++) {
				int lj = cdr3s.get(j).getCDR3length();
				// int dif = Math.abs(li - lj);
				int dif = 3 * Math.abs(li - lj);
				int alignment_length = Math.max(li, lj);
				// int alignment_length = li+lj;
				if (dif <= 0.32 * alignment_length) {
					if (!partitioned_set.contains(cdr3s.get(0))) {
						partitioned_set.add(cdr3s.get(0));
					}
					partitioned_set.add(cdr3s.get(j));
				}
			}
			//System.out.println("small group size:" + partitioned_set.size());
			if (partitioned_set.size() > 0) {
				partitioned.add(partitioned_set);
				cdr3s.removeAll(partitioned_set);
			} else {
				cdr3s.remove(cdr3s.get(0));
			}
			//System.out.println("remaining:" + cdr3s.size());
		}
		return partitioned;
	}

	public StringBuffer paritionedSetReadset(String inputFilename,
			ArrayList<CDR3> cdr3s, StringBuffer sb) throws Exception {

		NewClonalRelate newclonal = new NewClonalRelate();

		String xmlpath = inputFilename.substring(0, inputFilename
				.lastIndexOf("."))
				+ ".dendrogram.xml";
		Folder folder = newclonal.loadTree(xmlpath);

		// Use the below code to get maximum disance of a dataset.
		/*
		 * 
		 * List<String> ids = new ArrayList<String>();
		 * 
		 * //ids.add("5"); ids.add("27"); ids.add("20"); ids.add("19"); for (int
		 * i = 0; i < 90; i++) { // Integer PW57 = seqs.size() - 20 + i;
		 * ids.add(String.valueOf(seqs.size() - 90 + i)); } // Get the distance
		 * that needs to perform cutting float dis = getMaxDistance(folder,
		 * ids); System.out.println(dis);
		 */

		// Distance threshold can be changed here.
		float dis = 0.32f;

		List<Folder> subtrees = newclonal.cutFolder(folder, dis);
		File path = new File(xmlpath.substring(0, xmlpath.lastIndexOf("."))
				+ ".subtrees");
		if (!path.exists())
			path.mkdir();
		String pathString = path.getAbsolutePath();

		List<SortCluster> clusters = new ArrayList<SortCluster>();
		HashMap<String, Integer> seq_ids = new HashMap<String, Integer>();
		for (int i = 0; i < cdr3s.size(); i++) {
			seq_ids.put(cdr3s.get(i).getSeqID(), i);
		}
		// System.out.println(seq_ids);
		for (int i = 0; i < subtrees.size(); i++) {
			cluster_id += 1;
			Folder f = subtrees.get(i);
			NodeClassFilter classFilter = new NodeClassFilter(Leaf.NODE_TYPE);
			List<Node> leafnodes = f.extract(classFilter, true);
			List<String> seqs = new ArrayList<String>();
			for (Node leafnode : leafnodes) {
				if (leafnode instanceof Leaf) {
					Leaf leaf = (Leaf) leafnode;
					String leafId = leaf.get_id();
					// System.out.println(leafId);
					if (leafId != null) {
						try {
							// seq = Integer.parseInt(leafId);
							seqs.add(leafId);
						} catch (Exception e) {
							e.printStackTrace();
						}
					}
				}
			}
			// System.out.println(seqs);
			for (int n = 0; n < seqs.size(); n++) {
				int seqInt = seq_ids.get(seqs.get(n));
				String seqString = String.valueOf(cluster_id) + ";"
						+ String.valueOf(leafnodes.size()) + ";"
						+ cdr3s.get(seqInt).getCDR3()
						+ cdr3s.get(seqInt).getCDR3length() + ";"
						+ cdr3s.get(seqInt).getOriginal();
				// + lines.get(seqInt).toString();

				Integer clusterId = i;
				SortCluster clusterResult = new SortCluster(seqString,
						clusterId);
				clusters.add(clusterResult);
			}

			// System.out.println(f.getId());
			FileRootNode rootNode = new FileRootNode("root");
			rootNode.makeLogicRoot();
			rootNode.addChildNode(f);
			JDOMPersistenceMgr mgr = JDOMPersistenceMgr.getInstance();
			String filename = pathString + "/" + String.valueOf(i + 1) + ".xml";
			// System.out.println(filename);
			mgr.save(rootNode, filename);
			NewickConverter newick = new NewickConverter();
			newick.converter(filename);
		}
		//System.out.println(cluster_id);
		newclonal.sortCluster(clusters);
		sb = partitionedOutput(clusters, sb);
		return sb;
	}

	public StringBuffer partitionedOutput(List<SortCluster> clusters,
			StringBuffer sb) {
		// StringBuffer sb = new StringBuffer();
		for (int i = 0; i < clusters.size(); i++) {
			SortCluster clust = clusters.get(i);
			sb.append(clust.getSequenceId());
			// if (i < (clusters.size() - 1)) {
			sb.append("\n");
			// }
			// System.out.println(sb);
		}
		// System.out.println(subtrees.size());
		return sb;
	}
}
