package clustering;

import java.util.ArrayList;
import java.util.List;

import com.giantleap.common.nodebean.node.base.filter.NodeClassFilter;
import com.giantleap.common.nodebean.node.base.storage.node.Node;

import clustering.node.Folder;
import clustering.node.Leaf;

// Generate tree using average linkage

public class UPGMATreeGenerator {

	public static Folder newgenerateTree(Float[][] distanceMatrix,
			ArrayList<CDR3> cdr3s) {
		Folder[] folders = new Folder[distanceMatrix.length + 1];
		int[] DataIds = new int[cdr3s.size()];
		for (int i = 0; i < cdr3s.size(); i++)
			DataIds[i] = i;
		// DataIds[names.size()-1] = names.size()-1;
		int min_i = -1, min_j = -1;
		int k = 0;
		while (distanceMatrix.length > 0) {
			float min_dist = Float.MAX_VALUE;
			min_i = -1;
			min_j = -1;
			// Find the minimum value in the matrix
			for (int i = 0; i < distanceMatrix.length; i++) {
				for (int j = 0; j < distanceMatrix.length - i; j++) {
					if (min_dist > distanceMatrix[i][j]) {
						min_i = i;
						min_j = j;
						min_dist = distanceMatrix[i][j];
					}
				}
			}
			/*
			for (int i = 0; i < distanceMatrix.length; i++) {
				
				for (int j = 0; j < distanceMatrix[i].length; j++) {
					System.out.print(distanceMatrix[i][j]+",");
				}
				System.out.println();
			}*/
			int max_pos = min_i + min_j + 1;
			int num_leaf1 = 0;
			int num_leaf2 = 0;
			// Update the tree;
			if (folders[min_i] == null && folders[max_pos] == null) {
				/*
				 * IF combining two leaves, then a new node is created the
				 * distance between two leaves is distanceMatrix[min_i][min_j]
				 */
				num_leaf1 = 1;
				num_leaf2 = 1;
				Folder folder = new Folder();
				folder.makeLogicRoot();
				folder.set_id("-1");
				folder.set_step(k);
				folder.set_dist(distanceMatrix[min_i][min_j]);
				Leaf leaf_1 = new Leaf();
				Leaf leaf_2 = new Leaf();
				// leaf_1.set_id(String.valueOf(DataIds[min_i]));
				// leaf_2.set_id(String.valueOf(DataIds[max_pos]));
				leaf_1.set_id(String.valueOf(cdr3s.get(DataIds[min_i]).getSeqID()));
				leaf_2.set_id(String.valueOf(cdr3s.get(DataIds[max_pos]).getSeqID()));
				folder.addChildNode(leaf_1);
				folder.addChildNode(leaf_2);
				folders[min_i] = folder;
				// System.out.println(dDistanceMatrix[min_i][min_j]);
			} else if (folders[min_i] != null && folders[max_pos] != null) {
				/*
				 * IF combining two nodes, then a new node is created the
				 * distance between two nodes is distanceMatrix[min_i][min_j]
				 */
				num_leaf1 = countLeaves(folders[min_i]);
				num_leaf2 = countLeaves(folders[max_pos]);
				Folder folder = new Folder();
				folder.makeLogicRoot();
				folder.set_id("-1");
				folder.set_step(k);
				folder.set_dist(distanceMatrix[min_i][min_j]);

				Folder folder_min_i = folders[min_i];
				Folder folder_min_j = folders[max_pos];
				folder.addChildNode(folder_min_i);
				folder.addChildNode(folder_min_j);

				folders[min_i] = folder;

				// System.out.println(dDistanceMatrix[min_i][min_j]);
			} else if (folders[min_i] != null) {
				/*
				 * Combining a leaf and a node min_i is a node and min_j is a
				 * leaf
				 */
				num_leaf1 = countLeaves(folders[min_i]);
				num_leaf2 = 1;
				Folder folder = new Folder();
				folder.makeLogicRoot();
				folder.set_id("-1");
				folder.set_step(k);
				folder.set_dist(distanceMatrix[min_i][min_j]);
				Folder folder_min_i = folders[min_i];
				if (folder_min_i.getParent() != null) {
					folder_min_i.getParent().removeChildNode(folder_min_i,
							false);
				}

				folder.addChildNode(folder_min_i);
				Leaf leaf_1 = new Leaf();
				// leaf_1.set_id(String.valueOf(DataIds[min_j]));
				leaf_1.set_id(String.valueOf(cdr3s.get(DataIds[max_pos]).getSeqID()));
				folder.addChildNode(leaf_1);
				folders[min_i] = folder;
				// System.out.println(dDistanceMatrix[min_i][min_j]);

			} else {
				/*
				 * Combining a leaf and a node min_j is a node and min_i is a
				 * leaf
				 */
				num_leaf1 = 1;
				num_leaf2 = countLeaves(folders[max_pos]);
				Folder folder = new Folder();
				folder.makeLogicRoot();
				folder.set_id("-1");
				folder.set_step(k);
				folder.set_dist(distanceMatrix[min_i][min_j]);
				Folder folder_min_j = folders[max_pos];
				if (folder_min_j.getParent() != null) {
					folder_min_j.getParent().removeChildNode(folder_min_j,
							false);
				}

				folder.addChildNode(folder_min_j);
				Leaf leaf_1 = new Leaf();
				// leaf_1.set_id(String.valueOf(DataIds[min_i]));
				leaf_1.set_id(String.valueOf(cdr3s.get(DataIds[min_i]).getSeqID()));
				folder.addChildNode(leaf_1);
				folders[min_i] = folder;
				// System.out.println(dDistanceMatrix[min_i][min_j]);

			}
			//System.out.println(num_leaf1 + "      " + num_leaf2);
			for (int i = max_pos + 1; i < folders.length; i++) {
				folders[i - 1] = folders[i];
				// XMLTree[i-1] = XMLTree[i];
				DataIds[i - 1] = DataIds[i];
			}

			// Update distance matrix
			Float[][] tempDM = updateDM(distanceMatrix, min_i, min_j,
					num_leaf1, num_leaf2);
			distanceMatrix = tempDM;
			k++;
		}
		return folders[0];
	}

	public static Float[][] updateDM(Float[][] currentDM, int min_i, int min_j,
			int num_leaf1, int num_leaf2) {

		Float[][] tempDM = new Float[currentDM.length - 1][];
		int max_pos = min_i + min_j + 1;
		for (int i = 0; i < tempDM.length; i++) {
			Float[] tempDM_row = new Float[tempDM.length - i];
			for (int j = 0; j < tempDM.length - i; j++) {
				if (i == min_i) {
					if ((j + 1) > min_j) {
						float left_length = currentDM[i][j + 1];
						float right_length = currentDM[max_pos][j - min_j];
						tempDM_row[j] = (left_length * num_leaf1 + right_length
								* num_leaf2)
								/ (num_leaf1 + num_leaf2);
					}
					if (j < min_j) {
						float left_length = currentDM[i][j];
						float right_length = currentDM[j + i + 1][min_j - j - 1];
						tempDM_row[j] = (left_length * num_leaf1 + right_length
								* num_leaf2)
								/ (num_leaf1 + num_leaf2);
					}
				} else if (i != min_i) {
					if (i > (max_pos - 1)) {
						tempDM_row[j] = currentDM[i + 1][j];

					} else if ((i + j + 1) < (min_i + min_j)) {
						tempDM_row[j] = currentDM[i][j];
					} else if ((i + j + 1) > (min_i + min_j)) {
						tempDM_row[j] = currentDM[i][j + 1];
					} else if (i + j + 1 == min_i + min_j) {
						float left_length = currentDM[i][j];
						float right_length = currentDM[i][max_pos - 1 - i];
						tempDM_row[j] = (left_length * num_leaf1 + right_length
								* num_leaf2)
								/ (num_leaf1 + num_leaf2);
					}

				}
			}
			tempDM[i] = tempDM_row;
		}

		return tempDM;

	}

	public static int countLeaves(Folder f) {
		NodeClassFilter classFilter = new NodeClassFilter(Leaf.NODE_TYPE);
		List<Node> leafnodes = f.extract(classFilter, true);
		int num_leaf = leafnodes.size();
		return num_leaf;
	}
}
